<?php

namespace GGM\ProductManagerImplementation\ProductManagerMagento2;

use Magento\Backend\Model\Auth\Session;
use Magento\Framework\App\Config\ScopeConfigInterface;
use Magento\Framework\Exception\AuthenticationException;
use Magento\User\Model\User;
use stdClass;

/**
 * Class Magento2Router
 */
class Magento2Router
{
    const BRIDGE_VERSION_M2 = '2.3.8';

    /**
     * @var string
     */
    private $username;

    /**
     * @var ScopeConfigInterface
     */
    protected $scopeConfig;

    /**
     * @var ProductManagerConfigHelper
     */
    protected $productManagerConfigHelper;

    /**
     * @var ProductManagerDatabaseConnectionHelper
     */
    protected $productManagerDatabaseConnectionHelper;

    /**
     * @var ProductManagerImageHelper
     */
    protected $productManagerImageHelper;

    /**
     * @var ProductManagerReindexHelper
     */
    protected $productManagerReindexHelper;

    /**
     * @var OpenInMagentoHelper
     */
    protected $openInMagentoHelper;

    /**
     * @var User
     */
    protected $userModel;

    /**
     * @var Session
     */
    protected $authSession;

    /**
     * Magento2Router constructor.
     *
     * @param ScopeConfigInterface $scopeConfig
     * @param ProductManagerConfigHelper $productManagerConfigHelper
     * @param ProductManagerDatabaseConnectionHelper $productManagerDatabaseConnectionHelper
     * @param ProductManagerImageHelper $productManagerImageHelper
     * @param ProductManagerReindexHelper $productManagerReindexHelper
     * @param OpenInMagentoHelper $openInMagentoHelper
     * @param User $userModel
     * @param Session $authSession
     */
    public function __construct(
        ScopeConfigInterface $scopeConfig,
        ProductManagerConfigHelper $productManagerConfigHelper,
        ProductManagerDatabaseConnectionHelper $productManagerDatabaseConnectionHelper,
        ProductManagerImageHelper $productManagerImageHelper,
        ProductManagerReindexHelper $productManagerReindexHelper,
        OpenInMagentoHelper $openInMagentoHelper,
        User $userModel,
        Session $authSession
    ) {
        $this->scopeConfig = $scopeConfig;
        $this->productManagerConfigHelper = $productManagerConfigHelper;
        $this->productManagerDatabaseConnectionHelper = $productManagerDatabaseConnectionHelper;
        $this->productManagerImageHelper = $productManagerImageHelper;
        $this->productManagerReindexHelper = $productManagerReindexHelper;
        $this->openInMagentoHelper = $openInMagentoHelper;
        $this->userModel = $userModel;
        $this->authSession = $authSession;
    }

    /**
     * @param string $username
     *
     * @return $this
     */
    public function setUsername(string $username)
    {
        $this->username = $username;

        return $this;
    }

    /**
     * @return string
     */
    public function getUsername()
    {
        return $this->username;
    }

    /**
     * @param string $encryptionKey
     * @param string $username
     *
     * @return void
     *
     * @throws AuthenticationException
     */
    public function execute()
    {
        if (isset($_GET['editproduct'])) {
            $this->openInMagentoHelper
                ->openInMagento(
                    'editproduct',
                    $_GET['editproduct'],
                    null
                );
        } elseif (isset($_GET['editcategory'])) {
            $this->openInMagentoHelper
                ->openInMagento(
                    'editcategory',
                    $_GET['editcategory'],
                    null
                );
        } elseif (isset($_GET['viewproduct'])) {
            $this->openInMagentoHelper
                ->openInMagento(
                    'viewproduct',
                    $_GET['viewproduct'],
                    $_GET['storeid']
                );
        } elseif (isset($_GET['viewcategory'])) {
            $this->openInMagentoHelper
                ->openInMagento(
                    'viewcategory',
                    $_GET['viewcategory'],
                    $_GET['storeid']
                );
        } else {
            $encryptionKey = $this->scopeConfig->getValue(
                'admin/product_manager/encryption_key'
            );

            $startTime = microtime(true);

            self::writeHeaderAndCookie();

            $jsonRpc = self::getJSONFromPOSTFields($encryptionKey);

            $this->setUsername((string)$jsonRpc[0]->username);
            $encryptedpassword = $jsonRpc[0]->encryptedpassword;

            $iv = base64_decode($_POST['iv']);

            $password = self::decipher(base64_decode($encryptedpassword),
                base64_decode($encryptionKey), $iv);

            $user = $this->userModel->loadByUsername(
                $this->getUsername()
            );
            if (!$user->verifyIdentity($password)) {
                $this->authSession->processLogout();
                self::exitOnFatalError('FATAL_ERROR_WRONG_USERNAME_OR_PASSWORD');
            }

            $jsonRpcResult = $this->executeJsonRpc($jsonRpc[1]);
            $this->productManagerDatabaseConnectionHelper->closeConnection();

            $jsonRpcResult->executionTime = microtime(true) - $startTime;
            self::writeJSON($jsonRpcResult);
        }
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeBatch(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $jsonRpcResult->result = array();

        $count = count($jsonRpc->params);

        for ($i = 0; $i < $count; ++$i) {
            $jsonRpcResult->result[$i] = $this->executeJsonRpc($jsonRpc->params[$i]);
        }

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeSqlQuery(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $sql = $jsonRpc->params[1];
        $binds = array();

        for ($i = 2; $i < count($jsonRpc->params); ++$i) {
            array_push($binds, $jsonRpc->params[$i]);
        }

        if (!isset($jsonRpc->saveLastInsertId)) {
            $jsonRpc->saveLastInsertId = false;
        }

        $this->productManagerDatabaseConnectionHelper
            ->pdoExecuteSqlQuery(
                $jsonRpcResult,
                $sql,
                $binds,
                $jsonRpc->saveLastInsertId
            );

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeDatabaseConnect(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $databaseAPI = $jsonRpc->params[0];
        $this->productManagerDatabaseConnectionHelper
            ->executeDatabaseConnection($jsonRpcResult, $databaseAPI);

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeUploadImage(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $type = $jsonRpc->params[0];
        $filename = $jsonRpc->params[1];
        $data = $jsonRpc->params[2];
        $lastModificationTime = $jsonRpc->params[3];
        $failIfFileExists = $jsonRpc->params[4];
        $useDispretionPath = true;

        if (count($jsonRpc->params) > 5) {
            $useDispretionPath = $jsonRpc->params[5];
        }

        $this->productManagerImageHelper->uploadImage(
            $jsonRpcResult,
            $type,
            $filename,
            $data,
            $lastModificationTime,
            $failIfFileExists,
            $useDispretionPath
        );

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeDeleteImage(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $type = $jsonRpc->params[0];
        $filename = $jsonRpc->params[1];
        $this->productManagerImageHelper
            ->deleteImage($jsonRpcResult, $type, $filename);

        return $this;
    }

    /**
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeGetConfig(stdClass $jsonRpcResult)
    {
        $this->productManagerConfigHelper
            ->getConfig($jsonRpcResult);

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeGetSourceModels(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $store_id = $jsonRpc->params[0];
        $locale_code = $jsonRpc->params[1];
        $this->productManagerConfigHelper
            ->getSourceModels($jsonRpcResult, $store_id, $locale_code);

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeLoadSaveProducts(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $productIds = $jsonRpc->params;
        $this->productManagerReindexHelper
            ->loadAndSaveProducts($jsonRpcResult, $productIds);

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeRegenerateProductRewrites(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $productIds = $jsonRpc->params;
        $this->productManagerReindexHelper
            ->regenerateProductsUrlRewrites($jsonRpcResult, $productIds);

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeReindexProducts(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $productIds = $jsonRpc->params->productIds;
        $reindexerIds = $jsonRpc->params->reindexerIds;
        $this->productManagerReindexHelper
            ->reindexProductsUsingIndexers(
                $jsonRpcResult,
                $productIds,
                $reindexerIds
            );

        return $this;
    }

    /**
     * @param stdClass $jsonRpc
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeCleanCache(stdClass $jsonRpc, stdClass $jsonRpcResult)
    {
        $productIds = $jsonRpc->params;
        $this->productManagerReindexHelper
            ->cleanProductsCache($jsonRpcResult, $productIds);

        return $this;
    }

    /**
     * @param stdClass $jsonRpcResult
     *
     * @return $this
     */
    public function executeConnect(stdClass $jsonRpcResult)
    {
        $jsonRpcResult->result = new stdClass();
        $jsonRpcResult->result->bridgeversion = self::BRIDGE_VERSION_M2;
        $jsonRpcResult->result->platform = 'Magento 2';
        $jsonRpcResult->result->bridgetype = 'Bridge file';
        $jsonRpcResult->result->bridgeapiversion = '2';

        return $this;
    }

    /**
     * @param $jsonRpc
     *
     * @return stdClass
     */
    public function executeJsonRpc($jsonRpc)
    {
        $startTime = microtime(true);
        $jsonRpcResult = new stdClass();

        switch ($jsonRpc->method) {
            case 'batch':
                $this->executeBatch($jsonRpc, $jsonRpcResult);
            break;
            case 'sqlquery':
                $this->executeSqlQuery($jsonRpc, $jsonRpcResult);
            break;
            case 'databaseconnection':
                $this->executeDatabaseConnect($jsonRpc, $jsonRpcResult);
            break;
            case 'uploadimage':
                $this->executeUploadImage($jsonRpc, $jsonRpcResult);
            break;
            case 'deleteimage':
                $this->executeDeleteImage($jsonRpc, $jsonRpcResult);
            break;
            case 'getconfig':
                $this->executeGetConfig($jsonRpcResult);
            break;
            case 'getsourcemodels':
                $this->executeGetSourceModels($jsonRpc, $jsonRpcResult);
            break;
            case 'loadandsaveproducts':
                $this->executeLoadSaveProducts($jsonRpc, $jsonRpcResult);
            break;
            case 'regenerateproductsurlrewrites':
                $this->executeRegenerateProductRewrites($jsonRpc, $jsonRpcResult);
            break;
            case 'reindexproductsusingindexers':
                $this->executeReindexProducts($jsonRpc, $jsonRpcResult);
            break;
            case 'cleanproductscache':
                $this->executeCleanCache($jsonRpc, $jsonRpcResult);
            break;
            case 'connect':
                $this->executeConnect($jsonRpcResult);
            break;
        }

        $jsonRpcResult->id = $jsonRpc->id;
        $jsonRpcResult->executionTime = microtime(true) - $startTime;

        return $jsonRpcResult;
    }

    /**
     * @param $code
     * @param string $reason
     *
     * @return void
     */
    public static function exitOnFatalError($code, $reason = '')
    {
        $result = new stdClass();
        //$result->result = null;

        if ($reason == '') {
            $result->error = $code;
        } else {
            $result->error = $code . ', ' . $reason;
        }

        echo json_encode($result);
        exit();
    }

    /**
     * @return void
     */
    public static function writeHeaderAndCookie()
    {
        if (isset($_SERVER)) {
            $_SERVER['SCRIPT_FILENAME'] = str_replace($_SERVER['SCRIPT_NAME'], '/index.php', $_SERVER['SCRIPT_FILENAME']);
            $_SERVER['REQUEST_URI'] = '/index.php';
            $_SERVER['SCRIPT_NAME'] = '/index.php';
            $_SERVER['PHP_SELF'] = '/index.php';
        }

        header($_SERVER['SERVER_PROTOCOL'] . ' 202 Accepted');  //200 OK
        header('Cache-Control: no-store, private, no-cache, must-revalidate');
        header('Cache-Control: pre-check=0, post-check=0, max-age=0, max-stale=0', false);
        header('Expires: Sat, 26 Jul 1997 05:00:00 GMT');
        header('Expires: 0', false);
        header('Last-Modified: ' . gmdate('D, d M Y H:i:s') . ' GMT');
        header('Pragma: no-cache');
        header('Content-type: application/json');
        setcookie('EXTERNAL_NO_CACHE', '1');
        setcookie('nocache', '1');
        setcookie('no-cache', '1');
        setcookie('NO_CACHE', '1');
        setcookie('external_no_cache', '1');
    }

    /**
     * @param $data
     * @param $key
     * @param $iv
     *
     * @return false|string
     */
    public static function decipher($data, $key, $iv)
    {
        if (function_exists('openssl_decrypt')) {
            $plaintext = openssl_decrypt($data, 'aes-256-cbc', $key, OPENSSL_RAW_DATA, $iv);
        } else {
            // @codingStandardsIgnoreStart
            $td = @mcrypt_module_open(MCRYPT_RIJNDAEL_128, '', MCRYPT_MODE_CBC, '');
            @mcrypt_generic_init($td, $key, $iv);
            $plaintext = @mdecrypt_generic($td, $data);
            // @codingStandardsIgnoreEnd

            $last = substr($plaintext, -1);
            $plaintext = substr($plaintext, 0, strlen($plaintext) - ord($last));

            // @codingStandardsIgnoreStart
            @mcrypt_generic_deinit($td);
            @mcrypt_module_close($td);
            // @codingStandardsIgnoreEnd
        }

        return $plaintext;
    }

    /**
     * @param $encryptionKey
     *
     * @return mixed|void
     */
    public static function getJSONFromPOSTFields($encryptionKey)
    {
        //check if all the fields are set
        if (!isset($_POST['c']) and !isset($_POST['iv'])) {
            self::exitOnFatalError('FATAL_ERROR_POST_FIELDS_MISSING');
        } else {
            //encrypted data
            if (function_exists('mcrypt_module_open')) {
                $iv = base64_decode($_POST['iv']);
                $cipherText = base64_decode($_POST['c']);
                $json = self::decipher($cipherText, base64_decode($encryptionKey), $iv);
                $jsonRpc = json_decode($json);

                if ($jsonRpc == null) {
                    self::exitOnFatalError('FATAL_ERROR_INVALID_JSON', $json);
                }
            } else {
                $json = $_POST['c'];
                $jsonRpc = json_decode($json);

                if ($jsonRpc == null) {
                    self::exitOnFatalError('FATAL_ERROR_INVALID_JSON_ENCRYPTION_NOT_SUPPORTED');
                }
            }

            return $jsonRpc;
        }

        self::exitOnFatalError('FATAL_ERROR_INVALID_JSON_ENCRYPTION_NOT_SUPPORTED');

        return new stdClass();
    }

    /**
     * @param $dataLength
     *
     * @return void
     */
    public static function printCompressedDataHeader($dataLength)
    {
        echo 'PM10';

        $a = ($dataLength << 24) & 0xFF;
        $b = ($dataLength << 16) & 0xFF;
        $c = ($dataLength << 8) & 0xFF;
        $d = $dataLength & 0xFF;

        echo chr($a) . chr($b) . chr($c) . chr($d);
    }

    /**
     * @param $jsonRpcResult
     *
     * @return void
     */
    public static function writeJSON($jsonRpcResult)
    {
        if (function_exists('gzcompress')) {
            $data = json_encode($jsonRpcResult);
            self::printCompressedDataHeader(strlen($data));
            echo gzcompress($data, 6);
        } else {
            echo json_encode($jsonRpcResult);
        }
    }
}